package com.eldos.idp;

import SecureBlackbox.Base.SBRandom;
import SecureBlackbox.Base.SBUtils;
import SecureBlackbox.Base.TElX509Certificate;
import SecureBlackbox.SAMLCommon.*;
import SecureBlackbox.SAMLIDP.TElSAMLServiceProviderInfo;
import SecureBlackbox.XML.SBXMLDefs;
import SecureBlackbox.XML.TElXMLDOMDocument;
import SecureBlackbox.XMLSecurity.TElXMLKeyInfoX509Data;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class Utils {

    private static final long MAX_ISSUE_INSTANT_TIME_DIFF = 60000;
    private static final int NOT_BEFORE_TIMEOUT = 60000;
    private static final int ASSERTIONS_TTL = 60000 * 30;
    private static final int SAML_SESSION_TTL = 60000 * 30;

    public static void processRequest(HttpServletRequest req, HttpServletResponse resp, TElXMLDOMDocument doc,
                                      String login, String state,
                                      List<TElSAMLServiceProviderInfo> spList, String prefferedSSOResponseBinding,
                                      String idpIssuer, String signCertPath,
                                      boolean signResponses) throws ServletException, IOException {
        if ( ! "AuthnRequest".equals(doc.getDocumentElement().getLocalName())) {
            resp.setStatus(400); // Bad request
            return;
        }

        TElSAMLAuthnRequestElement samlReq = new TElSAMLAuthnRequestElement();
        samlReq.loadFromXML(doc.getDocumentElement());

        // find SP
        TElSAMLServiceProviderInfo currentSP = null;
        String id;

        // by issuer
        for (TElSAMLServiceProviderInfo spInfo : spList) {
            id = spInfo.getEntityID();
            if (id != null && id.equals(samlReq.getIssuer().getValue())) {
                currentSP = spInfo;
                break;
            }
        }
        // by sp name qualifier
        if (currentSP == null) {
            for (TElSAMLServiceProviderInfo spInfo : spList) {
                id = spInfo.getSPNameQualifier();
                if (id != null && id.equals(samlReq.getIssuer().getValue())) {
                    currentSP = spInfo;
                    break;
                }
            }
        }
        // by ACS URL
        String acsUrl = samlReq.getAssertionConsumerServiceURL();
        String location;

        if (currentSP == null && acsUrl != null) {
            for (TElSAMLServiceProviderInfo spInfo : spList) {
                for (int i = 0; i < spInfo.getAssertionConsumerServiceCount(); i++) {
                    location = spInfo.getAssertionConsumerServices(i).getLocation();
                    if (location != null && acsUrl.equals(location)) {
                        currentSP = spInfo;
                        break;
                    }
                }
            }
        }
        // by certificate
        final int keyCount = samlReq.getSigned() ? samlReq.getSignature().getKeyInfo().getCount() : 0;
        TElX509Certificate signCert;

        if (currentSP == null && keyCount > 0) {
            for (int i = 0; i < keyCount; i++) {
                signCert = ((TElXMLKeyInfoX509Data)samlReq.getSignature().getKeyInfo().getItem(i)).getCertificate();
                if (signCert != null) {
                    for (TElSAMLServiceProviderInfo spInfo : spList) {
                        if (signCert.equals(spInfo.getSigningCertificate())) {
                            currentSP = spInfo;
                            break;
                        }
                    }
                }
            }
        }
        // SP not found
        if (currentSP == null) {
            resp.setStatus(500); // internal error
            return;
        }

        // reload request with signature checking
        TElSAMLSignatureHandler sigHandler = new TElSAMLSignatureHandler();
        sigHandler.setCertificate(currentSP.getSigningCertificate());
        sigHandler.setValidateReferences(true);

        samlReq = new TElSAMLAuthnRequestElement();
        samlReq.setHandler(sigHandler);
        samlReq.loadFromXML(doc.getDocumentElement());

        // check request
        if ( ! SBXMLSAMLCore.SAMLVersion.equals(samlReq.getVersion())) {
            resp.setStatus(400); // Bad request
            return;
        }

        long diff = SBUtils.dateTimeToUnixTime(samlReq.getIssueInstant()) - SBUtils.dateTimeToUnixTime(SBUtils.utcNow());
        if (diff > MAX_ISSUE_INSTANT_TIME_DIFF) {
            resp.setStatus(400); // Bad request
            return;
        }

        acsUrl = samlReq.getAssertionConsumerServiceURL();
        int acsIdx = samlReq.getAssertionConsumerServiceIndex();
        short bindingId = -1;

        // get response destination URL
        if (acsIdx >= 0) {
            TElSAMLEndpoint ep = currentSP.findAssertionConsumerServiceEP(acsIdx);
            if (ep != null) {
                acsUrl = ep.getLocation();
                bindingId = ep.getBinding();
            } else {
                resp.setStatus(400); // Bad request
                return;
            }
        } else if (acsUrl != null && ( ! "".equals(acsUrl))) {
            ArrayList<?> epList = currentSP.findAssertionConsumerServiceEP(acsUrl);

            if (epList.size() == 1) {
                TElSAMLEndpoint ep = (TElSAMLEndpoint)epList.get(0);

                bindingId = ep.getBinding();
                acsUrl = ep.getLocation();

                String bindingStr = samlReq.getProtocolBinding();
                if (bindingStr != null && ( ! "".equals(bindingStr))) {
                    bindingId = TElSAMLBinding.strToBinding(bindingStr);
                }
            } else if (epList.size() > 1) {
                TElSAMLEndpoint ep = null;
                String bindingStr = samlReq.getProtocolBinding();

                if (bindingStr != null && ( ! "".equals(bindingStr)))
                    ep = findEPByBinding(epList, TElSAMLBinding.strToBinding(bindingStr));

                if (ep == null)
                    ep = findEPByBinding(epList, TElSAMLBinding.strToBinding(prefferedSSOResponseBinding));

                if (ep != null) {
                    bindingId = ep.getBinding();
                    acsUrl = ep.getLocation();
                } else {
                    resp.setStatus(400); // Bad request
                    return;
                }
            } else {
                resp.setStatus(400); // Bad request
                return;
            }
        } else {
            TElSAMLEndpoint ep = currentSP.findDefaultAssertionConsumerServiceEP();

            if (ep == null) {
                ep = currentSP.findAssertionConsumerServiceEP(TElSAMLBinding.strToBinding(prefferedSSOResponseBinding));

                if (ep == null && currentSP.getAssertionConsumerServiceCount() > 0) {
                    ep = currentSP.getAssertionConsumerServices(0);
                } else {
                    throw new RuntimeException("No endpoint to send response to!");
                }

                bindingId = ep.getBinding();
                acsUrl = ep.getLocation();
            }
        }

        // create response
        boolean useSSL = (req.getScheme() != null) && "https".equals(req.getScheme().toLowerCase());

        TElSAMLResponseElement samlResp = createSuccessResponse(acsUrl, samlReq.getID(), idpIssuer);
        generateAssertions(samlResp, acsUrl, login, samlReq.getID(), idpIssuer, currentSP, useSSL);
        signResponse(samlResp, signResponses, signCertPath);
        String respStr = wrapResponse(samlResp, acsUrl, bindingId, state);

        if (bindingId == SBXMLSAMLBind.sbtRedirect) {
            resp.setStatus(302); // redirect
            resp.setHeader("Location", respStr);
        } else if (bindingId == SBXMLSAMLBind.sbtPOST) {
            resp.setStatus(200); // OK
            resp.getWriter().println(respStr);
        }
    }

    private static TElSAMLResponseElement createSuccessResponse(String acsUrl, String reqID, String idpIssuer) {
        TElSAMLResponseElement samlResp = new TElSAMLResponseElement();

        samlResp.setVersion(SBXMLSAMLCore.SAMLVersion);
        samlResp.setID(SBRandom.sbRndString(20));
        samlResp.setIssueInstant(SBUtils.utcNow());
        samlResp.setDestination(acsUrl);
        samlResp.setInResponseTo(reqID);

        TElSAMLIssuerElement issuer = new TElSAMLIssuerElement();
        issuer.setValue(idpIssuer);
        issuer.setFormat("");
        samlResp.setIssuer(issuer);

        TElSAMLStatusCodeElement statusCode = new TElSAMLStatusCodeElement();
        statusCode.setValue(SBXMLSAMLProtocol.xmlSAMLStatusSuccess);

        TElSAMLStatusElement status = new TElSAMLStatusElement();
        status.setStatusCode(statusCode);
        samlResp.setStatus(status);

        return samlResp;
    }

    private static String wrapResponse(TElSAMLResponseElement samlResp, String acsUrl, short bindingId, String state) {
        String res = null;

        TElXMLDOMDocument doc = new TElXMLDOMDocument();
        doc.appendChild(samlResp.saveToXML(doc));

        if (bindingId == SBXMLSAMLBind.sbtRedirect) {
            // Redirect binding
            TElSAMLRedirectBinding binding = new TElSAMLRedirectBinding();
            binding.setURL(acsUrl);
            binding.setRelayState(state);
            binding.setSignatureAlg(SBXMLDefs.xmlSignatureAlgorithmRSASHA1);

            if (samlResp.getSigned())
                binding.setKeyMaterial(((TElSAMLSignatureHandler)samlResp.getHandler()).getCertificate().getKeyMaterial());

            res = binding.saveResponse(doc);
        } else if (bindingId == SBXMLSAMLBind.sbtPOST) {
            // POST binding
            TElSAMLPOSTBinding binding = new TElSAMLPOSTBinding();
            binding.setURL(acsUrl);
            binding.setRelayState(state);
            binding.setMode(TSBSAMLPOSTBindingMode.spmServer);

            res = binding.saveResponse(doc);
        } else
            throw new RuntimeException("Unknown binding!");

        return res;
    }

    private static void signResponse(TElSAMLResponseElement samlResp, boolean signResponses, String signCertPath) {
        if (signResponses) {
            TElX509Certificate signCert = new TElX509Certificate();
            signCert.loadFromFileAuto(signCertPath, "");

            TElSAMLSignatureHandler sigHandler = new TElSAMLSignatureHandler();
            sigHandler.setCertificate(signCert);

            samlResp.setSigned(true);
            samlResp.setHandler(sigHandler);
        }
    }

    private static void generateAssertions(TElSAMLResponseElement samlResp, String acsUrl,
                                    String login, String reqID, String idpIssuer, TElSAMLServiceProviderInfo currentSP,
                                    boolean sslUsed) {
        Date notBefore = SBUtils.dateTimeAddMillis(SBUtils.utcNow(), NOT_BEFORE_TIMEOUT);
        Date notOnOrAfter = SBUtils.dateTimeAddMillis(notBefore, ASSERTIONS_TTL);

        TElSAMLAssertionElement assertion = new TElSAMLAssertionElement();
        assertion.setVersion(SBXMLSAMLCore.SAMLVersion);
        assertion.setID(SBRandom.sbRndString(20));
        assertion.setIssueInstant(SBUtils.utcNow());

        TElSAMLIssuerElement issuer = new TElSAMLIssuerElement();
        issuer.setValue(idpIssuer);
        issuer.setFormat("");
        assertion.setIssuer(issuer);

        TElSAMLSubjectElement subject = new TElSAMLSubjectElement();
        assertion.setSubject(subject);

        TElSAMLNameIDElement nameId = new TElSAMLNameIDElement();
        subject.setID(nameId);
        nameId.setFormat(SBXMLSAMLCore.xmlSAMLNameIDFormatTransient);
        nameId.setValue(login);

        int idx = assertion.getSubject().addSubjectConfirmation();
        TElSAMLSubjectConfirmationElement confirmation = assertion.getSubject().getSubjectConfirmations(idx);
        confirmation.set_Method(SBXMLSAMLProtocol.xmlSAMLConfirmationMethodBearer);

        TElSAMLSubjectConfirmationDataElement confirmData = new TElSAMLSubjectConfirmationDataElement();
        confirmation.setSubjectConfirmationData(confirmData);
        confirmData.setRecipient(acsUrl);
        confirmData.setNotBefore(notBefore);
        confirmData.setNotOnOrAfter(notOnOrAfter);
        confirmData.setInResponseTo(reqID);
        confirmData.setAType("");
        confirmData.setAddress("");

        TElSAMLConditionsElement conditions = new TElSAMLConditionsElement();
        assertion.setConditions(conditions);
        conditions.setNotBefore(notBefore);
        conditions.setNotOnOrAfter(notOnOrAfter);

        String entityID = currentSP.getEntityID();
        if (entityID != null && entityID.length() > 0) {
            TElSAMLAudienceRestrictionElement audience = new TElSAMLAudienceRestrictionElement();
            idx = audience.addAudience();
            audience.getAudienceList(idx).setValue(entityID);
            assertion.getConditions().addCondition(audience);
        }

        TElSAMLAuthnStatementElement authn = new TElSAMLAuthnStatementElement();
        assertion.addStatement(authn);

        authn.setAuthnInstant(SBUtils.utcNow());
        authn.setSessionNotOnOrAfter(SBUtils.dateTimeAddMillis(SBUtils.utcNow(), SAML_SESSION_TTL));
        authn.setSessionIndex(SBRandom.sbRndString(20));

        TElSAMLAuthnContextElement authnCtx = new TElSAMLAuthnContextElement();
        authn.setAuthnContext(authnCtx);
        authnCtx.setChoice(TSBSAMLAuthnContextChoice.sacV1);
        authnCtx.setAuthnContextClassRef("");
        authnCtx.setAuthnContextDeclRef("");

        if (sslUsed)
            authnCtx.setAuthnContextClassRef(SBXMLSAMLProtocol.xmlSAMLAuthnContextClassPasswordProtectedTransport);
        else
            authnCtx.setAuthnContextClassRef(SBXMLSAMLProtocol.xmlSAMLAuthnContextClassPassword);

        samlResp.addAssertion(assertion);
    }

    private static TElSAMLEndpoint findEPByBinding(ArrayList<?> list, short binding) {
        TElSAMLEndpoint ep;
        int sz = list.size();

        for (int i = 0; i < sz; i++) {
            ep = (TElSAMLEndpoint)list.get(i);
            if (ep.getBinding() == binding)
                return ep;
        }

        return null;
    }
}
